import { autocompletion } from '@codemirror/autocomplete';
import { setDiagnostics } from '@codemirror/lint';
import { Facet } from '@codemirror/state';
import { hoverTooltip } from '@codemirror/tooltip';
import { ViewPlugin, EditorView } from '@codemirror/view';
import { WebSocketTransport, RequestManager, Client } from '@open-rpc/client-js';

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _createForOfIteratorHelper(o, allowArrayLike) {
  var it;

  if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) {
    if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
      if (it) o = it;
      var i = 0;

      var F = function () {};

      return {
        s: F,
        n: function () {
          if (i >= o.length) return {
            done: true
          };
          return {
            done: false,
            value: o[i++]
          };
        },
        e: function (e) {
          throw e;
        },
        f: F
      };
    }

    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var normalCompletion = true,
      didErr = false,
      err;
  return {
    s: function () {
      it = o[Symbol.iterator]();
    },
    n: function () {
      var step = it.next();
      normalCompletion = step.done;
      return step;
    },
    e: function (e) {
      didErr = true;
      err = e;
    },
    f: function () {
      try {
        if (!normalCompletion && it.return != null) it.return();
      } finally {
        if (didErr) throw err;
      }
    }
  };
}

var timeout = 10000;
var serverUri = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var rootUri = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var documentUri = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var languageId = Facet.define({
  combine: function combine(values) {
    return values.reduce(function (_, v) {
      return v;
    }, '');
  }
});
var changesDelay = 500;

var LanguageServerPlugin = /*#__PURE__*/function () {
  function LanguageServerPlugin(view) {
    var _this = this;

    _classCallCheck(this, LanguageServerPlugin);

    this.view = view;
    this.rootUri = this.view.state.facet(rootUri);
    this.documentUri = this.view.state.facet(documentUri);
    this.languageId = this.view.state.facet(languageId);
    this.documentVersion = 0;
    this.changesTimeout = null;
    this.transport = new WebSocketTransport(this.view.state.facet(serverUri));
    this.requestManager = new RequestManager([this.transport]);
    this.client = new Client(this.requestManager);
    this.client.onNotification(function (data) {
      _this.processNotification(data);
    });
    this.initialize({
      documentText: this.view.state.doc.toString()
    });
  }

  _createClass(LanguageServerPlugin, [{
    key: "update",
    value: function update(_ref) {
      var _this2 = this;

      var docChanged = _ref.docChanged;
      if (!docChanged) return;
      if (this.changesTimeout) clearTimeout(this.changesTimeout);
      this.changesTimeout = setTimeout(function () {
        _this2.sendChange({
          documentText: _this2.view.state.doc.toString()
        });
      }, changesDelay);
    }
  }, {
    key: "destroy",
    value: function destroy() {
      this.client.close();
    }
  }, {
    key: "initialize",
    value: function initialize(_ref2) {
      var _this3 = this;

      var documentText = _ref2.documentText;
      this.client.request({
        method: 'initialize',
        params: {
          capabilities: {
            textDocument: {
              hover: {
                dynamicRegistration: true,
                contentFormat: ['plaintext', 'markdown']
              },
              synchronization: {
                dynamicRegistration: true,
                willSave: false,
                didSave: false,
                willSaveWaitUntil: false
              },
              completion: {
                dynamicRegistration: true,
                completionItem: {
                  snippetSupport: false,
                  commitCharactersSupport: true,
                  documentationFormat: ['plaintext', 'markdown'],
                  deprecatedSupport: false,
                  preselectSupport: false
                },
                contextSupport: false
              },
              signatureHelp: {
                dynamicRegistration: true,
                signatureInformation: {
                  documentationFormat: ['plaintext', 'markdown']
                }
              },
              declaration: {
                dynamicRegistration: true,
                linkSupport: true
              },
              definition: {
                dynamicRegistration: true,
                linkSupport: true
              },
              typeDefinition: {
                dynamicRegistration: true,
                linkSupport: true
              },
              implementation: {
                dynamicRegistration: true,
                linkSupport: true
              }
            },
            workspace: {
              didChangeConfiguration: {
                dynamicRegistration: true
              }
            }
          },
          initializationOptions: null,
          processId: null,
          rootUri: this.rootUri,
          workspaceFolders: [{
            name: 'root',
            uri: this.rootUri
          }]
        }
      }, timeout * 3).then(function (_ref3) {
        var capabilities = _ref3.capabilities;
            _ref3.serverInfo;
        _this3.capabilities = capabilities;

        _this3.client.notify({
          method: 'initialized',
          params: {}
        });

        _this3.client.notify({
          method: 'textDocument/didOpen',
          params: {
            textDocument: {
              uri: _this3.documentUri,
              languageId: _this3.languageId,
              text: documentText,
              version: _this3.documentVersion
            }
          }
        });

        _this3.ready = true;
      });
    }
  }, {
    key: "sendChange",
    value: function sendChange(_ref4) {
      var documentText = _ref4.documentText;
      if (!this.ready) return;

      try {
        this.client.notify({
          method: 'textDocument/didChange',
          params: {
            textDocument: {
              uri: this.documentUri,
              version: this.documentVersion++
            },
            contentChanges: [{
              text: documentText
            }]
          }
        });
      } catch (e) {
        console.error(e);
      }
    }
  }, {
    key: "requestDiagnostics",
    value: function requestDiagnostics(view) {
      this.sendChange({
        documentText: view.state.doc.toString()
      });
    }
  }, {
    key: "requestHoverTooltip",
    value: function requestHoverTooltip(view, _ref5) {
      var _this4 = this;

      var line = _ref5.line,
          character = _ref5.character;
      if (!this.ready || !this.capabilities.hoverProvider) return null;
      this.sendChange({
        documentText: view.state.doc.toString()
      });
      return new Promise(function (fulfill, reject) {
        _this4.client.request({
          method: 'textDocument/hover',
          params: {
            textDocument: {
              uri: _this4.documentUri
            },
            position: {
              line: line,
              character: character
            }
          }
        }, timeout).then(function (result) {
          if (!result) return;
          var contents = result.contents,
              range = result.range;
          var pos = posToOffset(view.state.doc, {
            line: line,
            character: character
          });
          var end;

          if (range) {
            pos = posToOffset(view.state.doc, range.start);
            end = posToOffset(view.state.doc, range.end);
          }

          if (pos === null) return fulfill(null);
          var dom = document.createElement('div');
          dom.classList.add('documentation');
          dom.textContent = formatContents(contents);
          fulfill({
            pos: pos,
            end: end,
            create: function create(view) {
              return {
                dom: dom
              };
            },
            above: true
          });
        })["catch"](function (reason) {
          reject(reason);
        });
      });
    }
  }, {
    key: "requestCompletion",
    value: function requestCompletion(context, _ref6, _ref7) {
      var line = _ref6.line,
          character = _ref6.character;
      var triggerKind = _ref7.triggerKind,
          triggerCharacter = _ref7.triggerCharacter;
      if (!this.ready || !this.capabilities.completionProvider) return null;
      this.sendChange({
        documentText: context.state.doc.toString()
      });
      return this.client.request({
        method: 'textDocument/completion',
        params: {
          textDocument: {
            uri: this.documentUri
          },
          position: {
            line: line,
            character: character
          },
          context: {
            triggerKind: triggerKind,
            triggerCharacter: triggerCharacter
          }
        }
      }, timeout).then(function (result) {
        if (!result) return null;
        var options = result.items.map(function (_ref8) {
          var detail = _ref8.detail,
              label = _ref8.label,
              kind = _ref8.kind,
              textEdit = _ref8.textEdit,
              documentation = _ref8.documentation,
              sortText = _ref8.sortText,
              filterText = _ref8.filterText;
          var completion = {
            label: label,
            detail: detail,
            apply: textEdit ? textEdit.newText : label,
            type: {
              1: 'text',
              2: 'method',
              3: 'function',
              // 4: Constructor
              // 5: Field
              6: 'variable',
              7: 'class',
              8: 'interface',
              // 9: Module
              10: 'property',
              // 11: Unit
              // 12: Value
              13: 'enum',
              14: 'keyword',
              // 15: Snippet
              // 16: Color
              // 17: File
              // 18: Reference
              // 19: Folder
              // 20: EnumMember
              21: 'constant'
            }[kind],
            sortText: sortText ? sortText : label,
            filterText: filterText ? filterText : label
          };
          if (documentation) completion.info = formatContents(documentation);
          return completion;
        });

        var _prefixMatch = prefixMatch(options),
            _prefixMatch2 = _slicedToArray(_prefixMatch, 2);
            _prefixMatch2[0];
            var match = _prefixMatch2[1];

        var token = context.matchBefore(match);
        var pos = context.pos;

        if (token) {
          pos = token.from;
          var word = token.text.toLowerCase();

          if (/^\w+$/.test(word)) {
            options = options.filter(function (_ref9) {
              var filterText = _ref9.filterText;
              return filterText.toLowerCase().indexOf(word) === 0;
            }).sort(function (a, b) {
              switch (true) {
                case a.apply.indexOf(token.text) === 0 && b.apply.indexOf(token.text) !== 0:
                  return -1;

                case a.apply.indexOf(token.text) !== 0 && b.apply.indexOf(token.text) === 0:
                  return 1;
              }

              return 0;
            });
          }
        }

        return {
          from: pos,
          options: options
        };
      });
    }
  }, {
    key: "processNotification",
    value: function processNotification(_ref10) {
      var method = _ref10.method;

      try {
        switch (method) {
          case 'textDocument/publishDiagnostics':
            this.processDiagnostics.apply(this, arguments);
        }
      } catch (error) {
        console.error(error);
      }
    }
  }, {
    key: "processDiagnostics",
    value: function processDiagnostics(_ref11) {
      var _this5 = this;

      var params = _ref11.params;
      var diagnostics = params.diagnostics.map(function (_ref12) {
        var range = _ref12.range,
            message = _ref12.message,
            severity = _ref12.severity;
        return {
          from: posToOffset(_this5.view.state.doc, range.start),
          to: posToOffset(_this5.view.state.doc, range.end),
          severity: {
            1: 'error',
            2: 'warning',
            3: 'info',
            4: 'info'
          }[severity],
          message: message
        };
      }).filter(function (_ref13) {
        var from = _ref13.from,
            to = _ref13.to;
        return from !== null && to !== null;
      }).sort(function (a, b) {
        switch (true) {
          case a.from < b.from:
            return -1;

          case a.from > b.from:
            return 1;
        }

        return 0;
      });
      this.view.dispatch(setDiagnostics(this.view.state, diagnostics));
    }
  }]);

  return LanguageServerPlugin;
}();
function languageServer(options) {
  var plugin;
  return [serverUri.of(options.serverUri), rootUri.of(options.rootUri), documentUri.of(options.documentUri), languageId.of(options.languageId), ViewPlugin.define(function (view) {
    return plugin = new LanguageServerPlugin(view);
  }), hoverTooltip(function (view, pos, side) {
    var _plugin;

    return (_plugin = plugin) === null || _plugin === void 0 ? void 0 : _plugin.requestHoverTooltip(view, offsetToPos(view.state.doc, pos));
  }), autocompletion({
    override: [function (context) {
      if (plugin == null) return null;
      var state = context.state,
          pos = context.pos,
          explicit = context.explicit;
      var line = state.doc.lineAt(pos);
      var trigKind = 1; // Invoked

      var trigChar = void 0;

      if (!explicit && plugin.capabilities && plugin.capabilities.completionProvider && plugin.capabilities.completionProvider.triggerCharacters && plugin.capabilities.completionProvider.triggerCharacters.indexOf(line.text[pos - line.from - 1]) >= 0) {
        trigKind = 2; // TriggerCharacter

        trigChar = line.text[pos - line.from - 1];
      }

      if (trigKind === 1 && !context.matchBefore(/\w+$/)) return null;
      return plugin.requestCompletion(context, offsetToPos(state.doc, pos), {
        triggerKind: trigKind,
        triggerCharacter: trigChar
      });
    }]
  }), baseTheme];
}

function posToOffset(doc, pos) {
  if (pos.line >= doc.lines) return null;
  return doc.line(pos.line + 1).from + pos.character;
}

function offsetToPos(doc, offset) {
  var line = doc.lineAt(offset);
  return {
    line: line.number - 1,
    character: offset - line.from
  };
}

function formatContents(contents) {
  if (Array.isArray(contents)) {
    var text = '';

    var _iterator = _createForOfIteratorHelper(contents),
        _step;

    try {
      for (_iterator.s(); !(_step = _iterator.n()).done;) {
        var c = _step.value;
        text += formatContents(c) + '\n\n';
      }
    } catch (err) {
      _iterator.e(err);
    } finally {
      _iterator.f();
    }

    return text;
  }

  if (typeof contents === 'string') return contents;
  return contents.value;
}

function toSet(chars) {
  var flat, preamble, words;
  preamble = "";
  flat = Object.keys(chars).join('');
  words = /\w/.test(flat);

  if (words) {
    preamble += "\\w";
    flat = flat.replace(/\w/g, '');
  }

  return "[".concat(preamble).concat(flat.replace(/[^\w\s]/g, '\\$&'), "]");
}

function prefixMatch(options) {
  var i, source;
  var first = {};
  var rest = {};

  var _iterator2 = _createForOfIteratorHelper(options),
      _step2;

  try {
    for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
      var o = _step2.value;
      var apply = o.apply;
      first[apply[0]] = true;

      for (var i = 1; i < apply.length; i++) {
        rest[apply[i]] = true;
      }
    }
  } catch (err) {
    _iterator2.e(err);
  } finally {
    _iterator2.f();
  }

  source = toSet(first) + toSet(rest) + "*$";
  return [new RegExp("^" + source), new RegExp(source)];
}
var baseTheme = EditorView.baseTheme({
  '.cm-tooltip.documentation': {
    display: 'block',
    marginLeft: '0',
    padding: '3px 6px 3px 8px',
    borderLeft: '5px solid #999',
    whiteSpace: 'pre'
  },
  '.cm-tooltip.lint': {
    whiteSpace: 'pre'
  }
});

export { languageServer };
